/********************************************************************
PSWaterlevel.cpp
 This runs as a task, that connects to the water level information
 site, retrieves the data, processes it and stores the data in a deque.
 *******************************************************************/
#define V2_8_0  //Comment this out to run it on V2.7.5

#include "predef.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <basictypes.h>
#include <buffers.h>
#include <utils.h>
#include <iosys.h>
#include <dns.h>
#include <ip.h>
#include <tcp.h>
#include "PSWLevel.h"
#include <ctime>
#include <deque>

#define DEST_PORT 	(80)
#define RX_BUFSIZE 	(8192)
#define DATA_BUFSIZE (131072) //the website data is daily water levels for the month
							  //it grows larger everyday until the start of the next month

extern "C" {
void GetDataTask(void *pdata);
}

DWORD GetDataTaskStack[USER_TASK_STK_SIZE*2] __attribute__( ( aligned( 4 ) ) );
struct DataPoint {long long unsigned time; float level;};
std::deque<DataPoint> PSWL;

//----- functions -----
void GetDataTask(void *pdata);
void GetErrorText(int code, size_t len, char* text);

/*-------------------------------------------------------------------
 Task to obtain data for plot
 -------------------------------------------------------------------*/
void InitializeGetDataTask()
{
	if (OSTaskCreate(GetDataTask, 
			NULL, 
	      (void *) &GetDataTaskStack[USER_TASK_STK_SIZE],
	      (void *) GetDataTaskStack,
	      MAIN_PRIO - 1) != OS_NO_ERR)
	{
		iprintf("*** Error creating GetData task\r\n");
	}
	else
		iprintf("Started GetDataTask\r\n");
}

/*-------------------------------------------------------------------
 Function to parse water level data and put it into a global deque
  Returns:
  0 on success
 -1 on failure
 -------------------------------------------------------------------*/
int ProcessData(char *source_str) {
	if (*source_str == '\0')
		return -1;
	int rval = 1;
	DataPoint PSData;
	float Height = 0;
	time_t t = time(NULL);
	tm* Date = localtime(&t);
	Date->tm_sec = 0; // must initialise seconds to zero
	char * pch;
	if (source_str != NULL) {
		PSWL.clear();

		pch = strtok(source_str, "\r\n"); // Move to the first data line
		iprintf(" %s\n", pch);
		while (rval > 0) {

			pch = strtok(NULL, " <div>/\r\n"); //Year
			if (pch == NULL) {
				rval = 0;
				break;
			}
			iprintf(" %s/", pch);
			Date->tm_year = atoi(pch) - 1900;

			pch = strtok(NULL, "/"); //Month
			if (pch == NULL)
				break;
			iprintf("%s/", pch);
			Date->tm_mon = atoi(pch) - 1;

			pch = strtok(NULL, ";"); //Day
			if (pch == NULL)
				break;
			iprintf("%s ", pch);
			Date->tm_mday = atoi(pch);

			pch = strtok(NULL, ":"); //Hour
			if (pch == NULL)
				break;
			iprintf("%s:", pch);
			Date->tm_hour = atoi(pch);

			pch = strtok(NULL, ";"); //Minute
			if (pch == NULL)
				break;
			iprintf("%s ", pch);
			Date->tm_min = atoi(pch);

			pch = strtok(NULL, "<;/\r\n"); //Height
			if (pch == NULL)
				break;
			iprintf("%5s = ", pch);
			Height = atof(pch);

			if (pch == NULL) {
				iprintf("At: %i", rval++);
				return 0;
			}
			PSData.time = mktime(Date);
			PSData.level = Height;
			PSWL.push_back(PSData);
			printf("Date: %lli, Level: %5.3f \r\n",PSData.time * 1000,
					PSData.level);
		}
	}

	return rval;
}     

/*-------------------------------------------------------------------
 Function to execute a DNS lookup to obtain the resultant IP address
 -------------------------------------------------------------------*/
void GetIPAddress(char *name, IPADDR* ip_addr)
{
	iprintf("Resolving URL: \"%s\"...\r\n", name);
#ifdef V2_8_0
	ip_addr->SetNull();
	int status = GetHostByName(name, ip_addr, ip_addr->NullIP(), TICKS_PER_SECOND * 10);
#else
	*ip_addr = 0;
	int status = GetHostByName(name, ip_addr, 0, TICKS_PER_SECOND * 10);
#endif
	switch (status)
	{
		case DNS_OK:
			iprintf("URL: \"%s\" resolved to ", name);
			ShowIP(*ip_addr);
			iprintf("\r\n");
			break;
		case DNS_TIMEOUT:
			iprintf("*** DNS ERROR: TIMEOUT!\r\n");
			break;
		case DNS_NOSUCHNAME:
			iprintf("*** DNS ERROR: NO SUCH NAME!\r\n");
			break;
		case DNS_ERR:
			iprintf("*** DNS ERROR: RESPONSE ERROR!\r\n");
			break;
	}
   return;
}

/*-------------------------------------------------------------------
 Task to obtain water level data from website
 -------------------------------------------------------------------*/
//Global Variables //
char RxBuffer[RX_BUFSIZE];
char DataBuffer[DATA_BUFSIZE];
char DestName[] = "tides.gc.ca";
char DestGet[2048];
bool success = false;
IPADDR DestIP_Addr;
int fd, x;
DWORD BytesRead = 0;

void GetDataTask(void *pdata) {

	char start_str[] = "<div class=\"stationTextData\">";
	char finish_str[] = "<!-- MainContentEnd -->";
	char ERROR[128];
	char* pch;
	int Countdown, status, successes = 0, failures = 0;
	int TotalTime;
	unsigned int DataBufSz;
	bool Header;
	// Set up Client Request
	siprintf(DestGet,
			"GET /eng/station/month/?type=1&sid=11375&tz=EST&pres=2 HTTP/1.1\r\n"
					"Accept: text/html\r\n"
					"User-Agent: Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko\r\n"
					"Host: www.tides.gc.ca\r\n"
					"\r\n");

	while (1) { //Main Loop
		iprintf("\n----------------------------------------Start of Loop ----------------\n");
		Header = true;
		success = false;
		close(fd);
		fd = 0;
		// Convert destination name to an IP address using DNS
#ifdef V2_8_0
		while (DestIP_Addr.IsNull()) {
#else
		while (DestIP_Addr == 0) {
#endif
			GetIPAddress(DestName, &DestIP_Addr);
			OSTimeDly(TICKS_PER_SECOND);
		}
		// Connect to the destination IP Address
		 while (!success){
			fd = connect(DestIP_Addr, 0, DEST_PORT, TICKS_PER_SECOND * 5);
			if (fd <= 0) {
				GetErrorText(fd, 127, ERROR);
				iprintf("Error connecting to: ");
				ShowIP(DestIP_Addr);
				iprintf(":%d >>%s\r\n", DEST_PORT, ERROR);
				OSTimeDly(TICKS_PER_SECOND*5);
			} else {
				success = true;
// Comment out the next line to see it work without out-of-order buffers
				SetOutOfOrderbuffers(fd, 8);
			}
		}

		// Send out a GET request and store the received data
		if (success) {
			// Send GET request
			int n = write(fd, DestGet, strlen(DestGet));
			iprintf("\nSent %d bytes as GET request:\n%s", n, DestGet);
			DataBufSz = DATA_BUFSIZE - 1;
			TotalTime = Secs;

			// Loop and record all incoming data
			BytesRead = 0;
			DataBuffer[0] = '\0';
			while ((success) && (BytesRead < DataBufSz)) {
				Countdown = 30;
				while (dataavail(fd)==0) {
					OSTimeDly(4);
					if (--Countdown == 0) {
						success = false;
						iprintf("\n\rConnection timed out, n = %d\r\n", n);
						break;
					} else
						iprintf("BytesRead: %lu Countdown: %i Total Seconds: %lu    \r",
								BytesRead, Countdown, Secs-TotalTime);
				}
				if (success) {
					int n = SockReadWithTimeout(fd, RxBuffer, RX_BUFSIZE - 1,1);
					if (n > 0) {
						BytesRead += n;
						RxBuffer[n] = '\0';
						if (Header) {
							DataBufSz = GetContentLength(RxBuffer);
							iprintf("\n\rResponse Header (%d bytes):\n%s\n\n",
									n, RxBuffer);
							Header = false;
						}
						if (BytesRead > 15730) {// The first part of the website is static and uninteresting
							strcat(DataBuffer, RxBuffer);
						}
					} else { // connection closed by server or timeout
						success = false;
						iprintf("\n\rConnection closed by server;\r\n");
					}
					if (success == false)
						iprintf("1\r\n");
				}
				if (success == false)
					iprintf("2\r\n");
			}
			iprintf("3\r\n");
		}
		iprintf("*** FD = %i closing.\r\n",fd);
		iprintf("*** FD Closed Reporting: %i \r\n",close(fd));
		fd = 0;

		if (success) { //Check for the Start Condition
			pch = strstr(DataBuffer, start_str);
			if (pch != NULL) { //Check for the Finish Condition
				char* fpch = strstr(pch, finish_str);
				if (fpch != NULL) {
					*fpch = 0;//terminate the string at the end of data position
					iprintf("Finish String found at position %lu\r\n", strlen(pch));
					status = 0;
//					status = ProcessData(pch);  // this is the parser; commented out because it just clogs the mtty window
				} else {
					success = false;
					iprintf("Didn't find Finish string in %lu bytes\r\n",
							strlen(pch));
					status = -1;
				}
			} else {
				success = false;
				iprintf("Didn't find Start string in %lu bytes\r\n", strlen(pch));
				status = -2;
			}
			iprintf("BytesRead: %lu\r\n", BytesRead);

			if (success) { // no errors, so wait for next reading
				iprintf(">>> Success!\r\n");
				successes++;
				// Sleep until next update
				DWORD delay_time = Secs + 30; // 3600; // set to 30 for testing.
				iprintf("%i Successes; %i Failures; Last Time: %lu Seconds  \r\n", successes, failures,Secs-TotalTime);
				iprintf(
						"Countdown to next data acquisition: %i Seconds       \r",
						(int) (delay_time - Secs));
				while (Secs < delay_time) {
					OSTimeDly(TICKS_PER_SECOND);
					iprintf(
							"Countdown to next data acquisition: %i Seconds       \r",
							(int) (delay_time - Secs));
				}
			} else {
				iprintf("*** Error %i processing data, restarting process\r\n",
						status);
			}
		}
		if(!success){
			iprintf(">>> FAILURE! <<<<<<<<<<<<<<<\r\n");
			failures++;
			iprintf("%i Successes; %i Failures; Last Time: %lu Seconds  \r\n", successes, failures,Secs-TotalTime);
		}
		iprintf("\r\n------------------------------------------End of Loop ----------\r\n");
	} // End Main Loop
}

int GetContentLength(char* header){
	if(strstr(header,"OK")==NULL)
			return -1;
	char * pch = strstr(header,"Content-Length:");
	int i = atoi(pch+15);
	pch = strstr(pch,"\r\n\r\n");
	*pch=0;


	iprintf("Content-Length = %i    \r\n",i);
	return i;

}
void GetErrorText(int code, size_t len, char* text){
	switch (code) {
	case TCP_ERR_NORMAL:
		snprintf(text,len,"(0) --- Indicates that the connection failed");
		break;
	case TCP_ERR_TIMEOUT:
		snprintf(text,len,"(-1) --- Indicates that the connection has timed out");
		break;
	case TCP_ERR_NOCON:
		snprintf(text,len,"(-2) --- Indicates that you have attempted to read/write from a socket that does not have a connection established yet");
		break;
	case TCP_ERR_CLOSING:
		snprintf(text,len,"(-3) --- Indicates that you have attempted to read/write from a socket that has already been closed");
		break;
	case TCP_ERR_NOSUCH_SOCKET:
		snprintf(text,len,"(-4) --- Indicates that you have attempted to allocate a socket that does not exist");
		break;
	case TCP_ERR_NONE_AVAIL:
		snprintf(text,len,"(-5) --- Indicates that you have attempted to allocate a socket, but no socket it currently available");
		break;
	case TCP_ERR_CON_RESET:
		snprintf(text,len,"(-6) --- Indicates that you have attempted to read/write from a connection that has been reset by the other side");
		break;
	case TCP_ERR_CON_ABORT:
		snprintf(text,len,"(-7) --- This is an internal error that the client won't usually see");
		break;
	default:
		snprintf(text,len,"(%i) --- This is an valid file descriptor",code);
	}

}
